package kasapp

import (
	"context"
	"fmt"
	"log/slog"
	"net"
	"net/http"
	"os"
	"os/exec"
	"strings"
	"time"

	"github.com/bufbuild/protovalidate-go"
	"github.com/getsentry/sentry-go"
	"github.com/go-logr/logr"
	"github.com/spf13/cobra"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/cmd"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/module/modshared"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/grpctool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/logz"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/tlstool"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/pkg/kascfg"
	"go.opentelemetry.io/otel"
	"google.golang.org/grpc/grpclog"
	"google.golang.org/protobuf/encoding/protojson"
	"google.golang.org/protobuf/proto"
	"k8s.io/klog/v2"
	"sigs.k8s.io/yaml"
)

const (
	delayedShutdownGoroutineDumpInterval   = 15 * time.Minute
	delayedShutdownGoroutineDumpEnvVarName = "KAS_DEBUG_DELAYED_SHUTDOWN"
)

type App struct {
	ConfigurationFile string
}

func (a *App) Run(ctx context.Context) error {
	return cmd.RunAndLog(ctx, a.runGlobals)
}

func (a *App) runGlobals(ctx context.Context) (*slog.Logger, error) {
	cfg, err := LoadConfigurationFile(ctx, a.ConfigurationFile)
	if err != nil {
		return nil, err
	}
	ApplyDefaultsToKASConfigurationFile(cfg)
	v, err := protovalidate.New()
	if err != nil {
		return nil, err
	}
	err = v.Validate(cfg)
	if err != nil {
		return nil, fmt.Errorf("config validate: %w", err)
	}
	log, grpcLog, err := loggerFromConfig(cfg.Observability.Logging)
	if err != nil {
		return nil, err
	}
	// Sentry
	hub, err := constructSentryHub(cfg.Observability.Sentry.Dsn, cfg.Observability.Sentry.Environment)
	if err != nil {
		return log, fmt.Errorf("sentry: %w", err)
	}
	defer hub.Flush(time.Second)

	grpclog.SetLoggerV2(grpcLog) // pipe gRPC logs into slog
	logrLogger := logr.FromSlogHandler(log.Handler())
	// Kubernetes uses klog so here we pipe all logs from it to our logger via an adapter.
	klog.SetLogger(logrLogger)
	otel.SetLogger(logrLogger)
	otel.SetErrorHandler(&otelErrorHandler{
		log: log,
		hub: hub,
	})

	if os.Getenv(delayedShutdownGoroutineDumpEnvVarName) == "true" {
		context.AfterFunc(ctx, func() {
			log.Info("Starting to collect goroutine dumps to investigate delayed shutdown", slog.Int64("interval_sec", int64(delayedShutdownGoroutineDumpInterval.Seconds())))
			startStacktraceCollector(log, hub, delayedShutdownGoroutineDumpInterval)
		})
	}

	app := ConfiguredApp{
		Log:           log,
		Hub:           hub,
		Validator:     v,
		Configuration: cfg,
	}
	return log, app.Run(ctx)
}

// LoadConfigurationFile loads the configuration file.
// This function does not run any validation logic.
func LoadConfigurationFile(ctx context.Context, configFile string) (*kascfg.ConfigurationFile, error) {
	configYAML, err := os.ReadFile(configFile) //nolint: gosec
	if err != nil {
		return nil, fmt.Errorf("configuration file: %w", err)
	}
	cfg, err := yamlToConfig(configYAML)
	if err != nil {
		return nil, err
	}
	gen := cfg.GetConfig().GetCommand()
	if gen != "" {
		genCfg, genErr := runConfigCommand(ctx, cfg.Config)
		if genErr != nil {
			return nil, fmt.Errorf("config command: %w", genErr)
		}
		proto.Merge(cfg, genCfg)
	}
	return cfg, nil
}

func NewCommand() *cobra.Command {
	a := App{}
	c := &cobra.Command{
		Use:   "kas",
		Short: "GitLab Kubernetes Agent Server",
		Args:  cobra.NoArgs,
		RunE: func(cmd *cobra.Command, args []string) error {
			return a.Run(cmd.Context())
		},
		SilenceErrors: true,
		SilenceUsage:  true,
	}
	c.Flags().StringVar(&a.ConfigurationFile, "configuration-file", "", "Configuration file to use (YAML)")
	cobra.CheckErr(c.MarkFlagRequired("configuration-file"))

	return c
}

func loggerFromConfig(loggingCfg *kascfg.LoggingCF) (*slog.Logger, *grpctool.Logger, error) {
	var level, grpcLevel slog.Level
	err := level.UnmarshalText([]byte(loggingCfg.Level.String()))
	if err != nil {
		return nil, nil, err
	}
	err = grpcLevel.UnmarshalText([]byte(loggingCfg.GrpcLevel.String()))
	if err != nil {
		return nil, nil, err
	}
	lockedWriter := &logz.LockedWriter{
		Writer: os.Stderr,
	}
	log := slog.New(slog.NewJSONHandler(lockedWriter, &slog.HandlerOptions{
		Level: level,
	}))
	grpcLog := &grpctool.Logger{
		Handler: slog.NewJSONHandler(lockedWriter, &slog.HandlerOptions{
			Level: grpcLevel,
		}),
	}

	return log, grpcLog, nil
}

func runConfigCommand(ctx context.Context, cf *kascfg.ConfigCF) (*kascfg.ConfigurationFile, error) {
	cmd, args := splitCommand(cf.Command)
	genCmd := exec.CommandContext(ctx, cmd, args...) //nolint: gosec
	genCmd.Stderr = os.Stderr                        // pipe cmd's stderr into our own stderr
	genConfigYAML, err := genCmd.Output()
	if err != nil {
		return nil, err
	}
	return yamlToConfig(genConfigYAML)
}

func yamlToConfig(configYAML []byte) (*kascfg.ConfigurationFile, error) {
	configJSON, err := yaml.YAMLToJSON(configYAML)
	if err != nil {
		return nil, fmt.Errorf("YAMLToJSON: %w", err)
	}
	cfg := &kascfg.ConfigurationFile{}
	err = protojson.Unmarshal(configJSON, cfg)
	if err != nil {
		return nil, fmt.Errorf("protojson.Unmarshal: %w", err)
	}
	return cfg, nil
}

func splitCommand(cmd string) (string, []string) {
	cmdAndArgs := strings.Split(cmd, " ")
	return cmdAndArgs[0], cmdAndArgs[1:]
}

func constructSentryHub(dsn, environment string) (*sentry.Hub, error) {
	dialer := net.Dialer{
		Timeout:   30 * time.Second,
		KeepAlive: 30 * time.Second,
	}
	sentryClient, err := sentry.NewClient(sentry.ClientOptions{
		Dsn:         dsn, // empty DSN disables Sentry transport
		SampleRate:  1,   // no sampling
		Release:     cmd.Version,
		Environment: environment,
		HTTPTransport: &http.Transport{
			Proxy:                 http.ProxyFromEnvironment,
			DialContext:           dialer.DialContext,
			TLSClientConfig:       tlstool.ClientConfig(),
			MaxIdleConns:          10,
			IdleConnTimeout:       90 * time.Second,
			TLSHandshakeTimeout:   10 * time.Second,
			ResponseHeaderTimeout: 20 * time.Second,
			ForceAttemptHTTP2:     true,
		},
	})
	if err != nil {
		return nil, err
	}
	scope := sentry.NewScope()
	scope.SetTag("git_ref", cmd.GitRef)
	return sentry.NewHub(sentryClient, scope), nil
}

type otelErrorHandler struct {
	log *slog.Logger
	hub *sentry.Hub
}

func (h *otelErrorHandler) Handle(err error) {
	h.log.LogAttrs(context.Background(), slog.LevelError, "OpenTelemetry error", logz.Error(err))
	capture(context.Background(), h.hub, "", modshared.NoAgentID, "OpenTelemetry error", err, nil)
}
