package ioz

import (
	"bytes"
	"crypto/rand"
	"encoding/base64"
	"errors"
	"io"
	"os"
	"path/filepath"
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/testing/testlogger"
)

const (
	// 1025 is a value above 1024 that is not mem.IsBelowBufferPoolingThreshold().
	// See https://github.com/grpc/grpc-go/issues/7631.
	minReadSize = 1025
)

func TestLoadBase64Secret(t *testing.T) {
	testcases := []struct {
		name          string
		srcLen        int
		minLen        int
		maxLen        int
		expectedError string
	}{
		{
			name:   "exact len match",
			srcLen: 32,
			minLen: 32,
			maxLen: 32,
		},
		{
			name:   "min len match",
			srcLen: 32,
			minLen: 32,
			maxLen: 40,
		},
		{
			name:   "max len match",
			srcLen: 40,
			minLen: 32,
			maxLen: 40,
		},
		{
			name:   "below min len",
			srcLen: 10,
			minLen: 32,
			maxLen: 40,
			// TODO make this an error in 18.0 https://gitlab.com/gitlab-org/cluster-integration/gitlab-agent/-/issues/617
			//expectedError: "is 10 bytes long; allowed length is from 32 to 40 bytes",
		},
		{
			name:   "above max len",
			srcLen: 41,
			minLen: 32,
			maxLen: 40,
			// TODO make this an error in 18.0 https://gitlab.com/gitlab-org/cluster-integration/gitlab-agent/-/issues/617
			//expectedError: "is 41 bytes long; allowed length is from 32 to 40 bytes",
		},
	}

	for _, tc := range testcases {
		t.Run(tc.name, func(t *testing.T) {
			file := filepath.Join(t.TempDir(), tc.name)
			data := make([]byte, tc.srcLen)
			_, err := rand.Read(data)
			require.NoError(t, err)

			b64data := []byte(base64.StdEncoding.EncodeToString(data))

			err = os.WriteFile(file, b64data, 0644)
			require.NoError(t, err)

			secret, err := LoadBase64Secret(testlogger.New(t), file, tc.minLen, tc.maxLen)
			if tc.expectedError != "" {
				assert.ErrorContains(t, err, tc.expectedError)
			} else {
				require.NoError(t, err)
				assert.Equal(t, data, secret)
			}
		})
	}
}

func TestReadAll_Reads(t *testing.T) {
	testcases := []struct {
		name         string
		reads        []readStep
		expectedErr  string
		expectedBufs int
	}{
		{
			name: "EOF",
			reads: []readStep{
				{
					err: io.EOF,
				},
			},
		},
		{
			name: "data,EOF",
			reads: []readStep{
				{
					n: minReadSize,
				},
				{
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "data+EOF",
			reads: []readStep{
				{
					n:   minReadSize,
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "0,data+EOF",
			reads: []readStep{
				{},
				{
					n:   minReadSize,
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "0,data,EOF",
			reads: []readStep{
				{},
				{
					n: minReadSize,
				},
				{
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "data,data+EOF",
			reads: []readStep{
				{
					n: minReadSize,
				},
				{
					n:   minReadSize,
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "error",
			reads: []readStep{
				{
					err: errors.New("boom"),
				},
			},
			expectedErr: "boom",
		},
		{
			name: "data+error",
			reads: []readStep{
				{
					n:   minReadSize,
					err: errors.New("boom"),
				},
			},
			expectedErr: "boom",
		},
		{
			name: "data,data+error",
			reads: []readStep{
				{
					n: minReadSize,
				},
				{
					n:   minReadSize,
					err: errors.New("boom"),
				},
			},
			expectedErr: "boom",
		},
		{
			name: "data,data+EOF - whole buf",
			reads: []readStep{
				{
					n: minReadSize,
				},
				{
					n:   readAllBufSize - minReadSize,
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "data,data,EOF - whole buf",
			reads: []readStep{
				{
					n: minReadSize,
				},
				{
					n: readAllBufSize - minReadSize,
				},
				{
					err: io.EOF,
				},
			},
			expectedBufs: 1,
		},
		{
			name: "data,data,EOF - 2 bufs",
			reads: []readStep{
				{
					n: readAllBufSize,
				},
				{
					n: minReadSize,
				},
				{
					n: readAllBufSize - minReadSize,
				},
				{
					n: minReadSize,
				},
				{
					err: io.EOF,
				},
			},
			expectedBufs: 3,
		},
	}

	for _, tc := range testcases {
		t.Run(tc.name, func(t *testing.T) {
			pool := &testPool{
				allocated: make(map[*[]byte]struct{}),
			}
			r := &stepReader{
				reads: tc.reads,
			}
			data, err := ReadAllWithPool(r, pool)
			if tc.expectedErr != "" {
				assert.EqualError(t, err, tc.expectedErr)
			} else {
				require.NoError(t, err)
				assert.Equal(t, r.read, data.Materialize())
				assert.Len(t, data, tc.expectedBufs)
				for i := 0; i < len(data)-1; i++ { // all but last should be full buffers
					assert.Equal(t, readAllBufSize, data[i].Len())
				}
			}
			data.Free()
			assert.Empty(t, pool.allocated)
		})
	}
}

func TestReadAll_WriteTo(t *testing.T) {
	testcases := []struct {
		name string
		size int
	}{
		{
			name: "small",
			size: minReadSize,
		},
		{
			name: "exact size",
			size: readAllBufSize,
		},
		{
			name: "big",
			size: readAllBufSize * 3,
		},
	}
	for _, tc := range testcases {
		t.Run(tc.name, func(t *testing.T) {
			pool := &testPool{
				allocated: make(map[*[]byte]struct{}),
			}
			buf := make([]byte, tc.size)
			_, err := rand.Read(buf)
			require.NoError(t, err)
			r := bytes.NewBuffer(buf)
			data, err := ReadAllWithPool(r, pool)
			require.NoError(t, err)

			assert.Equal(t, buf, data.Materialize())

			data.Free()
			assert.Empty(t, pool.allocated)
		})
	}
}
