package agent

import (
	"context"
	"errors"
	"fmt"
	"log/slog"
	"time"

	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v17/internal/tool/logz"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/client-go/kubernetes"
)

const (
	// The default timeout for Trivy scan is 5 minutes, using a timout to ensure that this function doesn't get stuck perpetually.
	// Giving additional 5 minutes after Trivy scan is complete for reading the chaned configmaps and transmitting the vulnerability report.
	namespaceScannerTimeout = 10 * time.Minute
)

type NamespaceScanner interface {
	scan(ctx context.Context, scanLogger *slog.Logger, targetNamespace string, reporter Reporter, scanningManager ScanningManager) ([]string, error)
}

type namespaceScanner struct {
	kubeClientset             kubernetes.Interface
	gitlabAgentNamespace      string
	gitlabAgentServiceAccount string
	agentID                   int64
	ocsServiceAccountName     string
	trivyK8sWrapperImage      string
}

func newNamespaceScanner(
	kubeClientset kubernetes.Interface,
	gitlabAgentNamespace string,
	gitlabAgentServiceAccount string,
	agentID int64,
	ocsServiceAccountName string,
	trivyK8sWrapperImage string,
) *namespaceScanner {
	return &namespaceScanner{
		kubeClientset:             kubeClientset,
		gitlabAgentNamespace:      gitlabAgentNamespace,
		gitlabAgentServiceAccount: gitlabAgentServiceAccount,
		agentID:                   agentID,
		ocsServiceAccountName:     ocsServiceAccountName,
		trivyK8sWrapperImage:      trivyK8sWrapperImage,
	}
}

func (ns *namespaceScanner) scan(ctx context.Context, scanLogger *slog.Logger, targetNamespace string, reporter Reporter, manager ScanningManager) ([]string, error) {
	ctx, cancel := context.WithTimeout(ctx, namespaceScannerTimeout)
	defer cancel()

	var uuids []string
	podName := fmt.Sprintf("trivy-scan-%s", targetNamespace)
	scanLogger = scanLogger.With(logz.PodName(podName))

	if err := manager.deleteChainedConfigmaps(ctx, targetNamespace); err != nil {
		return nil, fmt.Errorf("could not delete configmaps, before deploying scanning pod: %w", err)
	}

	podSpec, errP := manager.getScanningPodSpecs(podName, targetNamespace, ns.ocsServiceAccountName, ns.trivyK8sWrapperImage)
	if errP != nil {
		return nil, fmt.Errorf("could not get specs for OCS Scanning Pod: %w", errP)
	}

	scanLogger.Debug("Deploying OCS Scanning Pod")
	if err := manager.deployScanningPod(ctx, podSpec, podName); err != nil {
		return nil, fmt.Errorf("could not deploy scanning pod: %w", err)
	}

	defer func() {
		scanLogger.Debug("Deleting OCS Scanning Pod")
		if err := manager.deleteScanningPod(podName); err != nil {
			scanLogger.Error("Error deleting Pod", logz.Error(err))
		}
	}()

	scanLogger.Debug("Start watcher for OCS Scanning Pod")
	watcher, err := manager.watchScanningPod(ctx, podName)
	if err != nil {
		return nil, fmt.Errorf("could not start watcher for OCS Scanning Pod: %w", err)
	}
	defer watcher.Stop()

	done := ctx.Done()
	for {
		select {
		case <-done:
			scanLogger.Debug("Stopping watcher as context canceled")
			return nil, ctx.Err()
		case event, ok := <-watcher.ResultChan():
			if !ok {
				return nil, errors.New("channel closed unexpectedly")
			}

			pod, ok := event.Object.(*corev1.Pod)
			if !ok {
				return nil, errors.New("watcher received unexpected object that is not a Pod")
			}

			scanLogger.Debug("pod status", logz.PodStatus(string(pod.Status.Phase)))

			switch pod.Status.Phase {
			case corev1.PodPending:
				for _, containerStatus := range pod.Status.ContainerStatuses {
					waitState := containerStatus.State.Waiting
					if waitState != nil {
						msg, reason := waitState.Message, waitState.Reason
						switch reason {
						case ErrImagePullBackOff, ErrImageInspect, ErrImagePull, ErrImageNeverPull, ErrInvalidImageName:
							return nil, fmt.Errorf("waiting on OCS Scanning Pod error: %s reason: %s", msg, reason)
						default:
							scanLogger.Info("Waiting on OCS Scanning Pod", logz.Msg(msg), logz.Reason(reason))
						}
					}
				}
			case corev1.PodFailed:
				scanLogger.Error("OCS Scanning Pod failed")
				scanLogger.Info("Deleting remaining chained configmaps")
				if err := manager.deleteChainedConfigmaps(ctx, targetNamespace); err != nil {
					return nil, fmt.Errorf("could not delete configmaps: %w", err)
				}

				if len(pod.Status.ContainerStatuses) != 1 {
					return nil, errors.New("OCS Scanning pod should have only one container")
				}
				if pod.Status.ContainerStatuses[0].State.Terminated != nil {
					return nil, manager.extractExitCodeError(pod.Status.ContainerStatuses[0].State.Terminated.ExitCode,
						pod.Status.ContainerStatuses[0].State.Terminated.Reason)
				}
				return nil, errors.New("OCS Scanning pod exited with an error. Could not retrieve an exit code")
			case corev1.PodSucceeded:
				scanLogger.Info("OCS Scanning Pod Succeeded")
				scanLogger.Info("Reading chained configmaps")
				defer func() {
					scanLogger.Info("Deleting chained configmaps")
					if err := manager.deleteChainedConfigmaps(ctx, targetNamespace); err != nil {
						scanLogger.Error("Could not delete ConfigMaps", logz.Error(err))
					}
				}()
				payload, trivyVersion, err := manager.readChainedConfigmaps(ctx, targetNamespace)
				if err != nil {
					return nil, fmt.Errorf("could not read chained ConfigMaps: %w", err)
				}

				payloads, err := manager.parseScanningPodPayload(payload, trivyVersion)
				if err != nil {
					return nil, fmt.Errorf("could not parse OCS scanning pod report: %w", err)
				}

				uuids, err = reporter.Transmit(ctx, payloads)
				if err != nil {
					return nil, fmt.Errorf("error transmitting vulnerability reports: %w", err)
				}
				scanLogger.Info("Transmitted vulnerabilities to Gitlab")
				return uuids, nil
			case corev1.PodRunning:
				// Pod is running, continue watching
				continue
			case corev1.PodUnknown:
				// Pod might still be running but unreachable, continue watching
				continue
			default:
				// Unhandled pod status, continue watching
				continue
			}
		}
	}
}
